import asyncio
import os
import pathlib
import re
import textwrap

import asyncpg
import nest_asyncio
import pytest
import rtoml
import typer.testing

import pogo_migrate.cli
from pogo_migrate import sql
from pogo_migrate.context import Context
from pogo_migrate.migration import Migration


@pytest.fixture(autouse=True, scope="session")
def _apply_nest_asyncio():
    loop = asyncio.get_event_loop_policy().new_event_loop()
    nest_asyncio.apply(loop)
    try:
        yield
    finally:
        loop.close()


@pytest.fixture(autouse=True)
def _clear_migration_tracking():
    # ClassVar remembers history across tests. Clear it every test.
    try:
        yield
    except:  # noqa: E722, S110
        pass
    finally:
        Migration._Migration__migrations = {}


@pytest.fixture
def postgres_dsn():
    return os.environ["POSTGRES_DSN"]


@pytest.fixture(autouse=True)
def cwd(tmp_path):
    orig = pathlib.Path.cwd()

    try:
        os.chdir(str(tmp_path))
        yield tmp_path
    finally:
        os.chdir(orig)


@pytest.fixture
def pyproject_factory(cwd):
    def factory():
        p = cwd / "pyproject.toml"

        with p.open("w") as f:
            f.write(
                rtoml.dumps(
                    {
                        "tool": {
                            "pogo": {
                                "migrations": "./migrations",
                                "database_config": "{POSTGRES_DSN}",
                            },
                        },
                    },
                ),
            )

    return factory


@pytest.fixture
def migrations(cwd):
    p = cwd / "migrations"
    p.mkdir()

    return p


@pytest.fixture
def migration_file_factory(migrations):
    def factory(mig_id, format_, contents):
        p = migrations / f"{mig_id}.{format_}"
        with p.open("w") as f:
            f.write(contents)

        return p

    return factory


@pytest.fixture(autouse=True)
async def db_session(postgres_dsn):
    conn = await asyncpg.connect(postgres_dsn)
    tr = conn.transaction()
    await tr.start()
    await sql.ensure_pogo_sync(conn)
    try:
        yield conn
    finally:
        await tr.rollback()


@pytest.fixture
def context():
    return Context(0)


class CliRunner(typer.testing.CliRunner):
    target = pogo_migrate.cli.app
    result = None

    def invoke(self, *args, **kwargs):
        result = super().invoke(self.target, *args, **kwargs)
        self.result = result
        if result.exception:
            if isinstance(result.exception, SystemExit):
                # The error is already properly handled. Print it and return.
                print(result.output)  # noqa: T201
            else:
                raise result.exception.with_traceback(result.exc_info[2])
        return self.result

    def _clean_output(self, text: str):
        output = text.encode("ascii", errors="ignore").decode()
        output = re.sub(r"\s+\n", "\n", output)
        return textwrap.dedent(output).strip()

    def assert_output(self, expected):
        assert self._clean_output(self.result.output) == self._clean_output(expected)


@pytest.fixture
def cli_runner():
    return CliRunner()



from __future__ import annotations

import io
import logging
import sys
import traceback
import typing as t
from dataclasses import dataclass, field
from enum import IntEnum

if sys.version_info < (3, 10):
    from typing_extensions import ParamSpec
else:  # pragma: no cover
    from typing import ParamSpec

import click


class Verbosity(IntEnum):
    """Verbosity levels."""

    quiet = 0
    verbose1 = 1
    verbose2 = 2
    verbose3 = 3


P = ParamSpec("P")

VERBOSITY = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


def setup_logging(verbose: int = 0) -> None:
    """Configure the logging."""
    logging.basicConfig(
        level=VERBOSITY.get(verbose, logging.DEBUG),
        format="%(message)s",
        datefmt="[%X]",
    )
    asyncio_logger = logging.getLogger("asyncio")
    asyncio_logger.disabled = True
    sqlglot_logger = logging.getLogger("sqlglot")
    sqlglot_logger.disabled = True
    root_logger = logging.getLogger("")
    root_logger.setLevel(VERBOSITY.get(verbose, logging.DEBUG))


@dataclass
class Context:
    """Global context class."""

    verbose: int = field(default=0)
    _indent: int = field(default=0)

    def __post_init__(self: t.Self) -> None:
        setup_logging(self.verbose)

    def _echo(self: t.Self, message: str, *args: P.args) -> None:
        """Echo to the console."""
        message = message % args
        click.echo(f"{'  ' * self._indent}{message}")

    def error(self: t.Self, message: str, *args: P.args) -> None:
        """Echo to the console."""
        self._echo(message, *args)

    def warning(self: t.Self, message: str, *args: P.args) -> None:
        """Echo to the console for -v."""
        if self.verbose > Verbosity.quiet:
            self._echo(message, *args)

    def info(self: t.Self, message: str, *args: P.args) -> None:
        """Echo to the console for -vv."""
        if self.verbose > Verbosity.verbose1:
            self._echo(message, *args)

    def debug(self: t.Self, message: str, *args: P.args) -> None:
        """Echo to the console for -vvv."""
        if self.verbose > Verbosity.verbose2:
            self._echo(message, *args)

    def stacktrace(self: t.Self) -> None:
        """Echo exceptions to console for -vvv."""
        if self.verbose > Verbosity.verbose2:
            t, v, tb = sys.exc_info()
            sio = io.StringIO()
            traceback.print_exception(t, v, tb, None, sio)
            s = sio.getvalue()
            # Clean up odd python 3.11, 3.12 formatting on mac
            s = s.replace("\n    ^^^^^^^^^^^^^^^^^^^^^^^^^^", "")
            sio.close()
            self._echo(s)



import contextlib
import itertools

import pytest

from pogo_migrate import topologicalsort


class TestTopologicalSort:
    def check(self, nodes, edges, expected):
        deps: dict[str, set[str]] = {}
        edges = edges.split() if edges else []
        for a, b in edges:
            deps.setdefault(a, set()).add(b)
        output = list(topologicalsort.topological_sort(nodes, deps))
        for a, b in edges:
            with contextlib.suppress(ValueError):
                assert output.index(a) > output.index(b)
        assert output == list(expected)

    def test_it_keeps_stable_order(self):
        for s in map(str, itertools.permutations("ABCD")):
            self.check(s, "", s)

    def test_it_sorts_topologically(self):
        # Single group at start
        self.check("ABCD", "BA", "ABCD")  # pragma: no-spell-check
        self.check("BACD", "BA", "ABCD")  # pragma: no-spell-check

        # Single group in middle start
        self.check("CABD", "BA", "CABD")  # pragma: no-spell-check
        self.check("CBAD", "BA", "CABD")  # pragma: no-spell-check

        # Extended group
        self.check("ABCD", "BA DA", "ABCD")  # pragma: no-spell-check
        self.check("DBCA", "BA DA", "CADB")  # pragma: no-spell-check

        # Non-connected groups
        self.check("ABCDEF", "BC DE", "ACBEDF")
        self.check("ADEBCF", "BC DE", "AEDCBF")
        self.check("ADEFBC", "BC DE", "AEDFCB")
        self.check("DBAFEC", "BC DE", "AFEDCB")  # pragma: no-spell-check

    def test_it_discards_missing_dependencies(self):
        self.check("ABCD", "CX XY", "ABCD")

    def test_it_catches_cycles(self):
        with pytest.raises(topologicalsort.CycleError):
            self.check("ABCD", "AA", "")
        with pytest.raises(topologicalsort.CycleError):
            self.check("ABCD", "AB BA", "")  # pragma: no-spell-check
        with pytest.raises(topologicalsort.CycleError):
            self.check("ABCD", "AB BC CB", "")
        with pytest.raises(topologicalsort.CycleError):
            self.check("ABCD", "AB BC CA", "")

    def test_it_handles_multiple_edges_to_the_same_node(self):
        self.check("ABCD", "BA CA DA", "ABCD")  # pragma: no-spell-check
        self.check("DCBA", "BA CA DA", "ADCB")  # pragma: no-spell-check

    def test_it_handles_multiple_edges_to_the_same_node2(self):
        #      A --> B
        #      |     ^
        #      v     |
        #      C --- +
        for input_order in itertools.permutations("ABC"):
            self.check(input_order, "BA CA BC", "ACB")  # pragma: no-spell-check

    def test_it_doesnt_modify_order_unnecessarily(self):
        self.check("ABC", "CA", "ABC")



from pathlib import Path
from textwrap import dedent

import pytest

from pogo_migrate import migration, squash


def test_remove_no_dependent(migration_file_factory, context):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    squash.remove(context, m, None)

    assert mp.exists() is False


def test_remove_with_backup(migration_file_factory, context):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    squash.remove(context, m, None, backup=True)

    assert mp.exists() is False
    assert Path(f"{mp}.bak").exists() is True


def test_remove_no_parent_with_dependent(migration_file_factory, context):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )
    mp2 = migration_file_factory(
        "20210101_02_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends: 20210101_01_rando-commit

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m2 = migration.Migration(mp2.stem, mp2, None)

    m.load()
    m2.load()

    squash.remove(context, m, m2)

    assert m2.depends == set()


def test_remove_python_no_parent_with_dependent(migration_file_factory, context):
    mp = migration_file_factory(
        "20210101_01_rando-migration-message",
        "py",
        dedent('''
        """
        migration message
        """
        __depends__ = []
        __transaction__ = False

        async def apply(db):
            await db.execute("CREATE TABLE table_one();")
            await db.execute("CREATE TABLE table_two();")

        async def rollback(db):
            await db.execute("DROP TABLE table_two;")
            await db.execute("DROP TABLE table_one;")
        '''),
    )
    mp2 = migration_file_factory(
        "20210101_02_rando-migration-message",
        "py",
        dedent('''
        """
        migration message
        """
        __depends__ = ["20210101_01_rando-migration-message"]
        __transaction__ = False

        async def apply(db):
            await db.execute("CREATE TABLE table_one();")
            await db.execute("CREATE TABLE table_two();")

        async def rollback(db):
            await db.execute("DROP TABLE table_two;")
            await db.execute("DROP TABLE table_one;")
        '''),
    )

    m = migration.Migration(mp.stem, mp, None)
    m2 = migration.Migration(mp2.stem, mp2, None)

    m.load()
    m2.load()

    squash.remove(context, m, m2)

    assert m2.depends == set()


def test_remove_with_parent_with_dependent(migration_file_factory, context):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )
    mp2 = migration_file_factory(
        "20210101_02_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends: 20210101_01_rando-commit

        -- migrate: apply
        -- migrate: rollback
        """),
    )
    mp3 = migration_file_factory(
        "20210101_03_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends: 20210101_02_rando-commit

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m2 = migration.Migration(mp2.stem, mp2, None)
    m3 = migration.Migration(mp3.stem, mp3, None)

    m.load()
    m2.load()
    m3.load()

    squash.remove(context, m2, m3)

    assert m3.depends == {m}


def test_write_short_circuits_for_file_squashed_to_self():
    latest = migration.Migration("mig-id", "path", None)
    new = squash.write({}, {}, latest, None, [latest.id])
    assert new is None


def test_squash_file_created(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    new = squash.write({}, {}, m, None, ["squash-1", "squash-2"])

    assert str(new.path) == f"{m.path}.squash"


def test_message_maintained(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    new = squash.write({}, {}, m, None, ["squash-1", "squash-2"])

    content = new.path.read_text()
    assert content.startswith("-- commit")


def test_depends_updated(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    new = squash.write({}, {}, m, "previous-1", ["squash-1", "squash-2"])

    content = new.path.read_text()
    assert "-- depends: previous-1" in content


def test_squashed_migrations_tracked(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    new = squash.write({}, {}, m, "previous-1", ["squash-1", "squash-2", m.id])

    content = new.path.read_text()
    assert "-- squashed: squash-1" in content
    assert "-- squashed: squash-2" in content


def test_apply_statements_stored_in_order(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    apply_statements = {
        "table1": ["statement1", "statement2"],
        "table2": ["statement3", "statement4"],
    }
    new = squash.write(apply_statements, {}, m, "previous-1", ["squash-1", "squash-2", m.id])

    content = new.path.read_text()
    assert (
        dedent("""
    -- migrate: apply

    -- Squash table1 statements.

    statement1

    statement2

    -- Squash table2 statements.

    statement3

    statement4

    -- migrate: rollback
    """)
        in content
    )


def test_apply_data_statements_last(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    apply_statements = {
        "table1": ["statement1"],
        "__data": ["update", "insert", "delete"],
    }
    new = squash.write(apply_statements, {}, m, "previous-1", ["squash-1", "squash-2", m.id])

    content = new.path.read_text()
    assert (
        dedent("""
    -- migrate: apply

    -- Squash table1 statements.

    statement1

    -- Squash data statements.

    update

    insert

    delete

    -- migrate: rollback
    """)
        in content
    )


def test_rollback_statements_reversed_from_discovery_order(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    # Rollback statements are found in file order, in reverse statement order
    """
    file1:
    ALTER table2 DROP COLUMN a;
    DROP table2;
    ALTER table1 DROP COLUMN e;
    DROP table1;

    file2:
    ALTER table2 DROP COLUMN c;
    ALTER table2 DROP COLUMN d;
    ALTER table1 DROP COLUMN f;
    ALTER table1 DROP COLUMN g;
    """

    rollback_statements = {
        "table1": [
            ["DROP table1;", "ALTER table1 DROP COLUMN e;"],
            ["ALTER table1 DROP COLUMN g;", "ALTER table1 DROP COLUMN f;"],
        ],
        "table2": [
            ["DROP table2;", "ALTER table2 DROP COLUMN a;"],
            ["ALTER table2 DROP COLUMN d;", "ALTER table2 DROP COLUMN c;"],
        ],
    }
    new = squash.write({}, rollback_statements, m, "previous-1", ["squash-1", "squash-2", m.id])

    content = new.path.read_text()
    # Rolback statements should be applied last discovered to first
    assert (
        dedent("""
    -- migrate: rollback

    -- Squash table2 statements.

    ALTER table2 DROP COLUMN c;

    ALTER table2 DROP COLUMN d;

    ALTER table2 DROP COLUMN a;

    DROP table2;

    -- Squash table1 statements.

    ALTER table1 DROP COLUMN f;

    ALTER table1 DROP COLUMN g;

    ALTER table1 DROP COLUMN e;

    DROP table1;
    """)
        in content
    )


def test_rollback_data_statements_first_reversed_from_discovery_order(migration_file_factory):
    mp = migration_file_factory(
        "20210101_01_rando-commit",
        "sql",
        dedent("""
        -- commit
        -- depends:

        -- migrate: apply
        -- migrate: rollback
        """),
    )

    m = migration.Migration(mp.stem, mp, None)
    m.load()

    # Rollback statements are found in file order, in reverse statement order
    """
    file1:
    UPDATE table1;
    DELETE FROM table1;
    DROP table1;

    file2:
    INSERT INTO table1;
    """

    rollback_statements = {
        "table1": [
            ["DROP table1;"],
        ],
        "__data": [
            ["DELETE FROM table1;", "UPDATE table1;"],
            ["INSERT INTO table1;"],
        ],
    }
    new = squash.write({}, rollback_statements, m, "previous-1", ["squash-1", "squash-2", m.id])

    content = new.path.read_text()
    # Rolback statements should be applied last discovered to first
    assert (
        dedent("""
    -- migrate: rollback

    -- Squash data statements.

    INSERT INTO table1;

    UPDATE table1;

    DELETE FROM table1;

    -- Squash table1 statements.

    DROP table1;
    """)
        in content
    )


@pytest.mark.parametrize(
    ("statement", "expected_type"),
    [
        ("CREATE TABLE tbl (id INT);", "CREATE"),
        ("ALTER TABLE tbl ADD COLUMN id INT;", "ALTER"),
        ("DROP TABLE tbl;", "DROP"),
        ("UPDATE tbl SET id = 1;", "UPDATE"),
        ("INSERT INTO TABLE tbl (id) VALUES (1);", "INSERT"),
        ("DELETE FROM TABLE tbl WHERE id = 1", "DELETE"),
    ],
)
def test_parse_sqlglot_type(statement, expected_type, context):
    parsed = squash.parse_sqlglot(context, statement)

    assert parsed.statement_type == expected_type


@pytest.mark.parametrize(
    ("statement", "expected_identifier"),
    [
        ("CREATE TABLE tbl (id INT);", "tbl"),
        ("CREATE TABLE IF NOT EXISTS tbl (id INT);", "tbl"),
        ("CREATE INDEX ix_table_id ON tbl (id);", "tbl"),
        ("CREATE INDEX ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE INDEX IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE INDEX CONCURRENTLY ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE INDEX CONCURRENTLY IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX CONCURRENTLY ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX CONCURRENTLY IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE AGGREGATE my_agg (numeric) ( SFUNC = numeric_mul, STYPE=numeric );", "my_agg"),
        ("CREATE AGGREGATE IF NOT EXISTS my_agg (numeric) ( SFUNC = numeric_mul, STYPE=numeric );", "my_agg"),
        ("CREATE EXTENSION pgcrypto;", "pgcrypto"),
        ("CREATE EXTENSION IF NOT EXISTS pgcrypto;", "pgcrypto"),
        ('ALTER TABLE "lock" ADD COLUMN id INT;', "lock"),
        ('ALTER TABLE "lock" ALTER COLUMN id INT;', "lock"),
        ("DROP TABLE tbl;", "tbl"),
        ("DROP TABLE IF EXISTS tbl;", "tbl"),
        ("DROP AGGREGATE my_agg (numeric);", "my_agg"),
        ("DROP AGGREGATE IF EXISTS my_agg (numeric);", "my_agg"),
        ("DROP EXTENSION IF EXISTS pgcrypto;", "pgcrypto"),
        ("DROP EXTENSION pgcrypto;", "pgcrypto"),
        ("DROP INDEX ix_table_id;", "ix_table_id"),
        ("DROP INDEX IF EXISTS ix_table_id;", "ix_table_id"),
        ("DROP INDEX CONCURRENTLY ix_table_id;", "ix_table_id"),
        ("DROP INDEX CONCURRENTLY IF EXISTS ix_table_id;", "ix_table_id"),
        ("UPDATE tbl SET id = 1;", None),
        ("INSERT INTO TABLE tbl (id) VALUES (1);", None),
        ("DELETE FROM TABLE tbl WHERE id = 1", None),
    ],
)
def test_parse_sqlglot_identifier(statement, expected_identifier, context):
    parsed = squash.parse_sqlglot(context, statement)

    assert parsed.identifier == expected_identifier


@pytest.mark.parametrize(
    ("statement"),
    [
        ("CREATE TABLE lock (id INT);"),
        ("ALTER TABLE lock ADD COLUMN id INT;"),
        ("DROP TABLE lock;"),
    ],
)
def test_parse_sqlglot_identifier_aborts_at_table_keyword(statement, context):
    # Original code would pick up column identifier.
    with pytest.raises(squash.ParseError, match="Expected table name but got lock."):
        squash.parse_sqlglot(context, statement)


@pytest.mark.parametrize(
    ("statement", "expected_identifier"),
    [
        ("CREATE TABLE tbl (id INT);", "tbl"),
        ("CREATE TABLE IF NOT EXISTS tbl (id INT);", "tbl"),
        ("CREATE INDEX ix_table_id ON tbl (id);", "tbl"),
        ("CREATE INDEX ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE INDEX IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE INDEX CONCURRENTLY ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE INDEX CONCURRENTLY IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX CONCURRENTLY ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE UNIQUE INDEX CONCURRENTLY IF NOT EXISTS ix_table_id ON tbl USING btree (id);", "tbl"),
        ("CREATE AGGREGATE my_agg (numeric) ( SFUNC = numeric_mul, STYPE=numeric );", "my_agg"),
        ("CREATE AGGREGATE IF NOT EXISTS my_agg (numeric) ( SFUNC = numeric_mul, STYPE=numeric );", "my_agg"),
        ("CREATE EXTENSION pgcrypto;", "pgcrypto"),
        ("CREATE EXTENSION IF NOT EXISTS pgcrypto;", "pgcrypto"),
        ('ALTER TABLE "lock" ADD COLUMN id INT;', "lock"),
        ('ALTER TABLE "lock" ALTER COLUMN id INT;', "lock"),
        ("DROP TABLE tbl;", "tbl"),
        ("DROP TABLE IF EXISTS tbl;", "tbl"),
        ("DROP AGGREGATE my_agg (numeric);", "my_agg"),
        ("DROP AGGREGATE IF EXISTS my_agg (numeric);", "my_agg"),
        ("DROP EXTENSION IF EXISTS pgcrypto;", "pgcrypto"),
        ("DROP EXTENSION pgcrypto;", "pgcrypto"),
        ("DROP INDEX ix_table_id;", "ix_table_id"),
        ("DROP INDEX IF EXISTS ix_table_id;", "ix_table_id"),
        ("DROP INDEX CONCURRENTLY ix_table_id;", "ix_table_id"),
        ("DROP INDEX CONCURRENTLY IF EXISTS ix_table_id;", "ix_table_id"),
        ("UPDATE tbl SET id = 1;", None),
        ("INSERT INTO TABLE tbl (id) VALUES (1);", None),
        ("DELETE FROM TABLE tbl WHERE id = 1", None),
    ],
)
def test_parse_identifier(statement, expected_identifier, context):
    parsed = squash.parse(context, statement)

    assert parsed.identifier == expected_identifier



from pathlib import Path
from unittest import mock

import pytest

from pogo_migrate.context import Context


@pytest.mark.parametrize("verbosity", [0, 1, 2, 3])
def test_verbosity(verbosity, monkeypatch):
    monkeypatch.setattr(Context, "_echo", mock.Mock())
    c = Context(verbosity)

    messages = [
        "error",
        "warning",
        "info",
        "debug",
    ]
    for message in messages:
        getattr(c, message)(message)

    assert c._echo.call_args_list == [mock.call(message) for message in messages[: verbosity + 1]]


def test_stacktrace(monkeypatch):
    monkeypatch.setattr(Context, "_echo", mock.Mock())
    c = Context(3)

    try:
        raise Exception("message")  # noqa: TRY002, EM101
    except:  # noqa: E722
        c.stacktrace()

    name = Path(__file__)
    assert c._echo.call_args == mock.call(
        f"""Traceback (most recent call last):
  File "{name}", line 31, in test_stacktrace
    raise Exception("message")  # noqa: TRY002, EM101
Exception: message
""",
    )


def test_stacktrace_quiet(monkeypatch):
    monkeypatch.setattr(Context, "_echo", mock.Mock())
    c = Context(2)

    try:
        raise Exception("message")  # noqa: TRY002, EM101
    except:  # noqa: E722
        c.stacktrace()

    assert c._echo.call_count == 0



from __future__ import annotations

import contextlib
import typing as t

from pogo_migrate import exceptions, sql
from pogo_migrate.migration import Migration, topological_sort

if t.TYPE_CHECKING:
    import asyncpg

    from pogo_migrate.config import Config
    from pogo_migrate.context import Context


@contextlib.asynccontextmanager
async def transaction(db: asyncpg.Connection, migration: Migration) -> None:
    tr = None
    if migration.use_transaction:
        tr = db.transaction()
        await tr.start()

    try:
        yield
    except Exception:
        if tr:
            await tr.rollback()
        raise
    else:
        if tr:
            await tr.commit()


async def apply(context: Context, config: Config, db: asyncpg.Connection) -> None:
    await sql.ensure_pogo_sync(db)
    migrations = await sql.read_migrations(config.migrations, db)
    migrations = topological_sort([m.load() for m in migrations])

    for migration in migrations:
        try:
            migration.load()
            if not migration.applied:
                context.warning("Applying %s", migration.id)
                async with transaction(db, migration):
                    await migration.apply(db)
                    await sql.migration_applied(db, migration.id, migration.hash)
        except Exception as e:  # noqa: PERF203
            msg = f"Failed to apply {migration.id}"
            raise exceptions.BadMigrationError(msg) from e


async def rollback(context: Context, config: Config, db: asyncpg.Connection, count: int | None = None) -> None:
    await sql.ensure_pogo_sync(db)
    migrations = await sql.read_migrations(config.migrations, db)
    migrations = reversed(list(topological_sort([m.load() for m in migrations])))

    i = 0
    for migration in migrations:
        try:
            migration.load()
            if migration.applied and (count is None or i < count):
                context.warning("Rolling back %s", migration.id)

                async with transaction(db, migration):
                    await migration.rollback(db)
                    await sql.migration_unapplied(db, migration.id)
                    i += 1
        except Exception as e:  # noqa: PERF203
            msg = f"Failed to rollback {migration.id}"
            raise exceptions.BadMigrationError(msg) from e



from __future__ import annotations

import typing as t

from pogo_migrate import config, migrate, sql
from pogo_migrate.context import Context

if t.TYPE_CHECKING:
    import asyncpg


async def apply(db: asyncpg.Connection | None = None) -> None:
    context = Context()
    c = config.load_config()
    if db is None:
        db = await sql.get_connection(c.database_dsn)
    await migrate.apply(context, c, db)


async def rollback(db: asyncpg.Connection | None = None) -> None:
    context = Context()
    c = config.load_config()
    if db is None:
        db = await sql.get_connection(c.database_dsn)
    await migrate.rollback(context, c, db)



from unittest import mock

import pytest

from pogo_migrate import config, context, testing
from tests.util import AsyncMock


@pytest.fixture(autouse=True)
def pyproject(pyproject_factory):
    return pyproject_factory()


async def test_apply(monkeypatch, db_session, cwd):
    monkeypatch.setattr(testing.migrate, "apply", AsyncMock())
    await testing.apply(db_session)

    assert testing.migrate.apply.call_args == mock.call(
        context.Context(),
        config.Config(
            root_directory=cwd,
            migrations=cwd / "migrations",
            database_config="{POSTGRES_DSN}",
        ),
        db_session,
    )


async def test_apply_loads_db(monkeypatch, db_session, cwd):
    monkeypatch.setattr(testing.migrate, "apply", AsyncMock())
    monkeypatch.setattr(testing.sql, "get_connection", AsyncMock(return_value=db_session))
    await testing.apply()

    assert testing.migrate.apply.call_args == mock.call(
        context.Context(),
        config.Config(
            root_directory=cwd,
            migrations=cwd / "migrations",
            database_config="{POSTGRES_DSN}",
        ),
        db_session,
    )


async def test_rollback(monkeypatch, db_session, cwd):
    monkeypatch.setattr(testing.migrate, "rollback", AsyncMock())
    await testing.rollback(db_session)

    assert testing.migrate.rollback.call_args == mock.call(
        context.Context(),
        config.Config(
            root_directory=cwd,
            migrations=cwd / "migrations",
            database_config="{POSTGRES_DSN}",
        ),
        db_session,
    )


async def test_rollback_loads_db(monkeypatch, db_session, cwd):
    monkeypatch.setattr(testing.migrate, "rollback", AsyncMock())
    monkeypatch.setattr(testing.sql, "get_connection", AsyncMock(return_value=db_session))
    await testing.rollback()

    assert testing.migrate.rollback.call_args == mock.call(
        context.Context(),
        config.Config(
            root_directory=cwd,
            migrations=cwd / "migrations",
            database_config="{POSTGRES_DSN}",
        ),
        db_session,
    )



from unittest import mock


class AsyncMock(mock.MagicMock):
    def __init__(self, *args, name=None, **kwargs) -> None:
        super().__init__(name, *args, **kwargs)
        self.__str__ = mock.MagicMock(return_value=name)

    async def __call__(self, *args, **kwargs):
        return super().__call__(*args, **kwargs)



from __future__ import annotations

import hashlib
import importlib.util
import re
import typing as t

import sqlparse

from pogo_migrate import exceptions, topologicalsort

if t.TYPE_CHECKING:
    from pathlib import Path

    import asyncpg


def strip_comments(statement: str) -> str:
    return "\n".join([line for line in statement.split("\n") if not line.startswith("--")])


def terminate_statements(statements: list[str]) -> list[str]:
    """Clean up last statement, if its missing a `;` it can cause issues in squashes."""
    if statements and statements[-1][-1] != ";":
        stmt = statements.pop()
        statements.append(f"{stmt};")

    return statements


def read_sql_migration(path: Path) -> tuple[str, t.Awaitable, t.Awaitable, bool]:
    """Read a sql migration.

    Parse the message, [depends], apply statements, and rollback statements.
    """
    with path.open() as f:
        contents = f.read()
        try:
            metadata, contents = contents.split("-- migrate: apply")
        except ValueError as e:
            msg = f"{path.name}: No '-- migrate: apply' found."
            raise exceptions.BadMigrationError(msg) from e

        m = re.match(r".*--(.*)\s-- depends:(.*)[\s]?", metadata.strip())

        if m is None:
            msg = f"{path.name}: No '-- depends:' or message found."
            raise exceptions.BadMigrationError(msg)

        message = m[1].strip()
        depends = m[2].strip()
        use_transaction = "-- transaction: false" not in metadata

        try:
            apply_content, rollback_content = contents.split("-- migrate: rollback")
        except ValueError as e:
            msg = f"{path.name}: No '-- migrate: rollback' found."
            raise exceptions.BadMigrationError(msg) from e

        apply_statements = terminate_statements(sqlparse.split(apply_content.strip()))

        async def apply(db):  # noqa: ANN001, ANN202
            for statement in apply_statements:
                # Skip comments
                statement_ = strip_comments(statement)
                if statement_:
                    await db.execute(statement_)

        rollback_statements = terminate_statements(sqlparse.split(rollback_content.strip()))

        async def rollback(db):  # noqa: ANN001, ANN202
            for statement in rollback_statements:
                # Skip comments
                statement_ = strip_comments(statement)
                if statement_:
                    await db.execute(statement_)

        return message, depends, apply, rollback, use_transaction, apply_statements, rollback_statements


class Migration:
    __migrations: t.ClassVar[dict[str, Migration]] = {}

    def __init__(self: t.Self, mig_id: str | None, path: str, applied_migrations: set[str] | None) -> None:
        applied_migrations = applied_migrations or set()
        self.id = mig_id
        self.path = path
        self.hash = hashlib.sha256(mig_id.encode("utf-8")).hexdigest() if mig_id else None
        self._use_transaction: bool = True
        self._doc: str | None = None
        self._depends: set[Migration] | None = None
        self._apply: t.Awaitable | None = None
        self._rollback: t.Awaitable | None = None
        self._applied = self.id in applied_migrations
        self.__migrations[self.id] = self

    def __repr__(self: t.Self) -> str:
        return self.id

    @property
    def applied(self: t.Self) -> bool:
        return self._applied

    @property
    def depends(self: t.Self) -> set[Migration]:
        return self._depends

    @property
    def use_transaction(self: t.Self) -> bool:
        return self._use_transaction

    @property
    def depends_ids(self: t.Self) -> set[str]:
        return {m.id for m in self._depends}

    @property
    def is_sql(self: t.Self) -> bool:
        return self.path.suffix == ".sql"

    async def apply(self: t.Self, db: asyncpg.Connection) -> None:
        await self._apply(db)

    async def rollback(self: t.Self, db: asyncpg.Connection) -> None:
        await self._rollback(db)

    def load(self: t.Self) -> Migration:
        depends_ = []
        if self.is_sql:
            message, depends, apply, rollback, in_transaction, _, _ = read_sql_migration(self.path)
            self._doc = message
            self._apply = apply
            self._rollback = rollback
            self._use_transaction = in_transaction
            depends_ = depends.split()
        else:
            spec = importlib.util.spec_from_file_location(
                str(self.path),
                str(self.path),
            )

            if spec:
                module = importlib.util.module_from_spec(spec)

            if spec and spec.loader:
                try:
                    spec.loader.exec_module(module)
                except Exception as e:
                    msg = f"Could not import migration from '{self.path.name}'"
                    raise exceptions.BadMigrationError(msg) from e
                self._doc = module.__doc__.strip()
                depends_ = module.__depends__
                self._apply = module.apply
                self._rollback = module.rollback
                self._use_transaction = getattr(module, "__transaction__", True)
            else:
                msg = f"Could not import migration from '{self.path.name}': ModuleSpec has no loader attached"
                raise exceptions.BadMigrationError(msg)

        self._depends = {self.__migrations.get(mig_id) for mig_id in depends_}
        if None in self._depends:
            msg = f"Could not resolve dependencies for '{self.path.name}'"
            raise exceptions.BadMigrationError(msg)

        return self

    @property
    def __doc__(self: t.Self) -> str:
        return self._doc


def topological_sort(migrations: t.Iterable[Migration]) -> t.Iterable[Migration]:
    migration_list = list(migrations)
    all_migrations = set(migration_list)
    dependency_graph = {m: (m.depends & all_migrations) for m in migration_list}
    try:
        return list(topologicalsort.topological_sort(migration_list, dependency_graph))
    except topologicalsort.CycleError as e:
        msg = "Circular dependencies among these migrations {}".format(
            ", ".join(m.id for m in e.args[1]),
        )
        raise exceptions.BadMigrationError(msg) from e



from __future__ import annotations

import dataclasses
import os
import typing as t
from pathlib import Path
from string import Formatter

import rtoml

from pogo_migrate import exceptions

CONFIG_FILENAME = "pyproject.toml"


@dataclasses.dataclass
class Config:
    root_directory: Path
    migrations: Path
    database_config: str

    @property
    def database_dsn(self: t.Self) -> str:
        try:
            format_kwargs = {
                k[1]: os.environ[k[1]] for k in Formatter().parse(self.database_config) if k[1] is not None
            }
        except KeyError as e:
            msg = f"Configured database_config env var {e!s} not set."
            raise exceptions.InvalidConfigurationError(msg) from e

        return self.database_config.format(**format_kwargs)

    @classmethod
    def from_dict(cls: type[Config], data: dict, root_directory: Path) -> Config:
        data["root_directory"] = root_directory
        data["migrations"] = root_directory / data["migrations"]
        return cls(**data)


def find_config() -> Path | None:
    """Find the closest config file in the cwd or a parent directory"""
    d = Path.cwd()
    while d != d.parent:
        path = d / CONFIG_FILENAME
        if path.is_file():
            return path
        d = d.parent
    return None


def load_config() -> Config:
    config = find_config()
    if config is None:
        msg = f"No configuration found, missing {CONFIG_FILENAME}, run 'pogo init ...'"
        raise exceptions.InvalidConfigurationError(msg)

    with config.open() as f:
        data = rtoml.load(f)

    if "tool" not in data or "pogo" not in data["tool"]:
        msg = "No configuration found, run 'pogo init ...'"
        raise exceptions.InvalidConfigurationError(msg)

    return Config.from_dict(data["tool"]["pogo"], config.parent)
